#include <avr/io.h>
#include <avr/interrupt.h>

#define DEBUG		0			// debug mode enable

#define F_OSC		16000000

#define L1_ON		PORTA |= 0x40		// SMD green LED on PCB
#define L1_OFF		PORTA &= ~0x40
#define L2_ON		PORTA |= 0x80		// SMD yellow LED on PCB
#define L2_OFF		PORTA &= ~0x80

#define LEDS_ON		PORTA |= 0xc0;		// both LEDs on
#define LEDS_OFF	PORTA &= ~0xc0;		// both LEDs off

#define SPK_ON		PORTB |= 0x10		// speaker
#define SPK_OFF		PORTB &= ~0x10

#define POC_ON		PORTD |= 0x80		// counter with display for debugging
#define POC_OFF		PORTD &= ~0x80

#define PEN_DOWN	PORTD |= 0x40		// plotter pen
#define PEN_UP		PORTD &= ~0x40

#define CTS_STOP	PORTD |= 0x04		// signal CTS (output on pin PD2) to the computer, L = AT Mega can recieve data from PC
#define CTS_RECIEVE	PORTD &= ~0x04

#define XON	17	// CTRL+Q / continue	// not used for now.... it wasn't reliable enough
#define XOFF	19	// CTRL+S / stop

#define KEY_DELAY		60		// delay for buttons on panel

#define DEFAULT_DELIMITER	'*'		// default end of string in command LB = label

// All following values strongly depends on hardware

#define	MOTOR_DEFAULT_DELAY	14		// 30 for SMR 300-300 = original Minigraf
						//  6 for Alfi Replica and power supply voltage 15V (too warm motor and driver L298)
						// 11 for Alfi Replica and power supply voltage 12V
						// 14 for Alfi Replica and power supply voltage 10V (slow pen magnet)

#define PEN_UP_DELAY	200			// delay for pen movements
#define PEN_DOWN_DELAY	200

#define POWERSAVE_TIMEOUT	4000		// number of cycles until motors will be disabled and pen moved up when plotter does nothing
						// this value is something around 2 seconds

#define PWM_TOP			800		// 16 000 000 / 800 = 20 000 ... is 20kHz ok?
#define PWM_PEN_MOTOR		799		// 100% perfect
#define PWM_PAPER_MOTOR		500		// 62.5% less noisy, still strong enough
						// 75% is still noisy
						// 50% / 12V / Alfi Minebea Astrosyn is loosing steps

#define PAGEWIDTH_STEPS		6500		// Alfi (200 step motors)		6500 steps
						// Minigraf Aritma (80 step motors) 	2000 steps (verify value !)
						
volatile unsigned char limit_switch_found = 1;

//============================================================================
// explicit function declarations

void usart_putc (unsigned char c);

//==== RUZNE =================================================================

int strcmp (const char * s1, const char * s2)
{
	const unsigned char *p1 = (const unsigned char *)s1;
	const unsigned char *p2 = (const unsigned char *)s2;

	while (*p1 && *p1 == *p2) ++p1, ++p2;

	return (*p1 > *p2) - (*p2  > *p1);
}

void delay (unsigned int ms)
{
	unsigned int i,j;
		for ( j=0; j<100; j++)
			for ( i=0; i<ms; i++);
}

void beepdelay (unsigned int ms)
{
	unsigned int i,j;
		for ( j=0; j<10; j++)
			for ( i=0; i<ms; i++);
}

void simplebeep (void)
{
	unsigned char n;

	for (n = 0; n < 2; n++)
	{
		SPK_ON;
		beepdelay (2);
		SPK_OFF;
		beepdelay (2);
	}
}

void beep (unsigned int cycles, unsigned int del)
{
	unsigned int n;

	for (n = 0; n < cycles; n++)
	{
		SPK_ON;
		beepdelay (del);
		SPK_OFF;
		beepdelay (del);
		
		// for 16MHz clock
		// 20 = 1.667kHz
		// 23 = 1.471kHz
		// 26 = 1.316kHz
		// 29 = 1.179kHz
	}
}

void debug_pulses (unsigned char pulses)
{
	// generate pulses on debug output
	unsigned char n;

	for (n = 0; n < pulses; n++)
	{
		POC_ON;
		beepdelay (1);
		POC_OFF;
		beepdelay (1);
	}
}

void start_effect ()
{
	LEDS_ON;
	beep (50,30);
	beep (50,10);
	delay (200);
	L1_OFF;
	delay (100);
	L2_OFF;
	delay (100);
}

//==== MOTORS ================================================================

volatile const unsigned char motor_tab_or [] = {0x01, 0x05, 0x04, 0x06, 0x02, 0x0A, 0x08, 0x09};
volatile const unsigned char motor_tab_and [] = {0xf1, 0xf5, 0xf4, 0xf6, 0xf2, 0xfA, 0xf8, 0xf9};

volatile unsigned char motor1_step;		// global state od stepper motor X
volatile unsigned char motor2_step;		// global state od stepper motor Y

volatile unsigned char motor_delay;		// delay between steps, should be dynamically configurable

volatile unsigned int powersave_counter;	// when zero, motors will be disabled


//	driver	wire	port
//------------------------------------------
//	A	yellow	bit 0	enable A
//	B	red	bit 1	enable A
//	C	white	bit 2	enable B
//	D	blue	bit 3	enable B

//  0001  0x01
//  0101  0x05
//  0100  0x04
//  0110  0x06
//  0010  0x02
//  1010  0x0A
//  1000  0x08
//  1001  0x09

void motors (int step_m1, int step_m2, int d)
{
	PORTA &= motor_tab_and [step_m1];
	PORTA |= motor_tab_or [step_m1];

	PORTB &= motor_tab_and [step_m2];
	PORTB |= motor_tab_or [step_m2];
	delay (d);
}

unsigned char motor_cw (unsigned char step)
{
	if (step >= 7) step = 0; else step++;
	return (step);
}

unsigned char motor_ccw (unsigned char step)
{
	if (step == 0) step = 7; else step--;
	return (step);
}

void motors_on (void)
{
	// enable motors and set propper PWM if needed
	OCR1A = PWM_PAPER_MOTOR;	// motor 1
	OCR1B = PWM_PEN_MOTOR;		// motor 2
}

void motors_off (void)
{
	// powersave and for lower temperature
	OCR1A = 1;
	OCR1B = 1;
}

void powersave_off (void)
{
	motors_on ();
	delay (motor_delay);
}

void powersave_on (void)
{
	powersave_counter = POWERSAVE_TIMEOUT;
	motors_off ();
}

void move_pen_up (unsigned int delay_cycles)
{
	PEN_UP;
	if (delay_cycles > 0) delay (delay_cycles);
}

void move_pen_down (unsigned int delay_cycles)
{
	PEN_DOWN;
	if (delay_cycles > 0) delay (delay_cycles);
}

//==== RS232 =================================================================

#define USART_MAX_BYTES		8
#define USART_BUFFER_SIZE	100		// 30 is enough, if data flow control works properly

volatile unsigned char usart_buffer [USART_BUFFER_SIZE + 1];
volatile unsigned char usart_index_in;
volatile unsigned char usart_index_out;
volatile unsigned char usart_bytes;

void set_baudrate (unsigned long int new_baudrate)
{
	unsigned int ubrr = 0;

	// U2X = 1 (assynchronous double speed mode), default 9600bps, MCU clock 16MHz
	switch (new_baudrate)
	{
		case 2400:
			ubrr = 832;
			break;
		case 4800:
			ubrr = 416;
			break;
		default:
		case 9600:
			ubrr = 207;
			break;
		case 19200:
			ubrr = 103;
			break;
		case 57600:
			ubrr = 34;
			break;
		case 115200:
			ubrr = 16;
			break;
	}
	UBRRH = (unsigned char) (ubrr >> 8);
	UBRRL = (unsigned char) (ubrr);
}

int initialize_usart()
{
	unsigned char i;

	set_baudrate (9600);

	// clear buffer
	for (i = 0; i < USART_BUFFER_SIZE; i++) usart_buffer [i] = 0;

	usart_index_in = 0;
	usart_index_out = 0;
	usart_bytes = 0;

	// set registers
	UCSRA = 0x22;	// 0010 0010 (RXC, TXC, UDRE, FE, DOR, PE, U2X, MPCM)
	UCSRB = (1 << RXEN) | (1 << TXEN) | (1 << RXCIE);	// enable receiver and transmitter, enable RX interrupt
	UCSRC = (1 << URSEL) | (1 << USBS) | (3 << UCSZ0);      // 8 data bit, 2 stop bits
	// U2X manual page 147
}

void usart_putc (unsigned char c)
{
	// send one character - wait until register for sending is not free
	while(!(UCSRA & (1 << UDRE)));
	UDR = c;                // posli znak
}

void usart_puts (char *s)
{
	// send whole string of characters, but don't receive data during that process
	while (*s)
	{
		usart_putc(*s);
		s++;
	}
}

unsigned char usart_getc(void)
{
	// get recieved character from USART register
	unsigned char byte;

	// get and return byte from buffer
	if (usart_bytes > 0)
	{
		byte = usart_buffer [usart_index_out];

		usart_index_out++;
		if (usart_index_out >= USART_BUFFER_SIZE) usart_index_out = 0;

		usart_bytes--;
			if (usart_bytes < USART_MAX_BYTES) CTS_RECIEVE; // usart_putc (XON);

		return (byte);
	}
	else
		return (0);
}

void debug_command (char *s, char command[], char signal)
{
	usart_puts (s);
	usart_putc (command[0]);
	usart_putc (command[1]);
	usart_putc ('\n');
	if (signal)
	{
		debug_pulses (1);
// 		beep (10,10);
	}
}

void debug_int (int i)
{
	char c;
	unsigned char leading_zeros;
	
	leading_zeros = 0;

	if (i < 0)
	{
		usart_putc ('-');
		i = -i;
	}
	
	c = '0';
	while (i > 9999) { c++; i-= 10000; }
	if (leading_zeros || c != '0') { usart_putc (c); leading_zeros = 1; }
	c = '0';
	while (i > 999) { c++; i-= 1000; }
	if (leading_zeros || c != '0') { usart_putc (c); leading_zeros = 1; }
	c = '0';
	while (i > 99) { c++; i-= 100; }
	if (leading_zeros || c != '0') { usart_putc (c); leading_zeros = 1; }
	c = '0';
	while (i > 9) { c++; i-= 10; }
	if (leading_zeros || c != '0') { usart_putc (c); leading_zeros = 1; }
	c = '0';
	while (i > 0) { c++; i-= 1; }
	usart_putc (c);
}

void debug_coordinates (int x, int y)
{
	usart_puts ("xy [");
	debug_int (x);
	usart_puts (", ");
	debug_int (y);
	usart_puts ("]\n");
}


ISR (USART_RXC_vect)
{
	// interrupt after byte was recieved into USART
	unsigned char a;
	a = UDR;

	usart_buffer [usart_index_in] = UDR;
	usart_bytes++;

	if (usart_bytes > USART_MAX_BYTES) CTS_STOP;				// if buffer is almost full, stop sending device
	if (usart_bytes > USART_BUFFER_SIZE) {						// this should never happen
		L2_ON;
		simplebeep();
	}

	usart_index_in++;
	if (usart_index_in >= USART_BUFFER_SIZE) usart_index_in = 0;
}

//============================================================================
// paper size A4 210x297mm - printable area cca 190x270mm
// orig motors - 8 steps per mm - 1520 x 2160 steps (7600 x 10800 in HPGL coortinates)
// new motors - 20 steps per mm - 3800 x 5400 steps (7600 x 10800 in HPGL coortinates)
// for comparison - XY4160 have limits 7400x10400 in HPGL coordinates, and 10 motor steps per mm
// values needs to be revided by the real physical limits

#define	PLOTER_MAX_PEN_DOWN	500;	// how long can be pen down (timeout in cycles, not miliseconds)
#define	PLOTER_MAX_X		7600;	// HPGL limits determined by dimmensions of working area
#define	PLOTER_MAX_Y		10800;

typedef enum {UP, DOWN} penstates;

volatile penstates pen_status;
volatile unsigned int pen_timeout;	// will be decremented until 0, then pen will go up
volatile int pen_x;			// real absolute coordinates, how far from borders of working area pen really is
volatile int pen_y;			// (should it be in motor steps or HPGL units?)

void plotter_pen_up (void)
{
	powersave_off ();
	// no delay, no pen movement if pen is already where should be
	if (pen_status == DOWN) move_pen_up (PEN_UP_DELAY);
	pen_status = UP;
	if (DEBUG) usart_puts ("-U-\n");
	pen_timeout = 0;
}

void plotter_pen_down (void)
{
	powersave_off ();
	// no delay, no pen movement if pen is already where should be
	if (pen_status == UP) move_pen_down (PEN_DOWN_DELAY);
	pen_status = DOWN;
	if (DEBUG) usart_puts ("-D-\n");
	pen_timeout = PLOTER_MAX_PEN_DOWN;
}

/*void plotter_pen_refresh (void)
{
	pen_timeout = PLOTER_MAX_PEN_DOWN;
	if (pen_status == DOWN) move_pen_down (PEN_UP_DELAY); else move_pen_up (PEN_DOWN_DELAY);
}*/

void plotter_set_coordinates (int x, int y)
{
	pen_x = x;
	pen_y = y;
}

void plot_low (int x1, int y1, int x2, int y2)
{
	int deltax;
	int deltay;
	int d,i,x,y;

	if (DEBUG) {
		usart_puts ("plot low - ");
		usart_puts ("x1=="); debug_int (x1);
		usart_puts (", x2=="); debug_int (x2);
		usart_puts (", y1=="); debug_int (y1);
		usart_puts (", y2=="); debug_int (y2);
		usart_putc ('\n');
	}

	if (y2 >= y1)
	{
		deltay = y2 - y1;
		i = 1;
		y = y1;
	} else {
		deltay = y1 - y2;
		i = -1;
		y = y1;
	}

	if (x2 >= x1)
		deltax = x2 - x1;
	else
		deltax = x1 - x2;

	d = deltay + deltay - deltax;

	if (x2 > x1) {
		if (DEBUG) usart_puts ("x2>x1\n");
		for (x = x1; x < x2; x++) {
 			motor2_step = motor_ccw (motor2_step);
			if (d > 0) {
				y += i;
				if (i > 0) motor1_step = motor_ccw (motor1_step); else motor1_step = motor_cw (motor1_step);
				d = d - deltax - deltax;
			}
			d = d + deltay + deltay;
 			motors (motor1_step, motor2_step, motor_delay);
			
// 			if (DEBUG) { debug_coordinates (x,y); }
		}
	} else {
		if (DEBUG) usart_puts ("x2<=x1\n");
		for (x = x1; x > x2; x--) {
 			motor2_step = motor_cw (motor2_step);
			if (d > 0) {
				y += i;
				if (i > 0) motor1_step = motor_ccw (motor1_step); else motor1_step = motor_cw (motor1_step);
				d = d - deltax - deltax;
			}
			d = d + deltay + deltay;
 			motors (motor1_step, motor2_step, motor_delay);

// 			if (DEBUG) { debug_coordinates (x,y); }
		}
	}
}

void plot_high (int x1, int y1, int x2, int y2)
{
	int deltax;
	int deltay;
	int d,i,x,y;

	if (DEBUG) {
		usart_puts ("plot high - ");
		usart_puts ("x1=="); debug_int (x1);
		usart_puts (", x2=="); debug_int (x2);
		usart_puts (", y1=="); debug_int (y1);
		usart_puts (", y2=="); debug_int (y2);
		usart_putc ('\n');
	}

	if (x2 >= x1)
	{
		deltax = x2 - x1;
		i = 1;
		x = x1;
	} else {
		deltax = x1 - x2;
		i = -1;
		x = x1;
	}

	if (y2 >= y1)
		deltay = y2 - y1;
	else
		deltay = y1 - y2;

	d = deltax + deltax - deltay;

	if (y2 >= y1) {
		if (DEBUG) usart_puts ("y2>y1\n");
		for (y = y1; y < y2; y++) {
 			motor1_step = motor_ccw (motor1_step);
			if (d > 0) {
				x += i;
				if (i > 0) motor2_step = motor_ccw (motor2_step); else motor2_step = motor_cw (motor2_step);
				d = d - deltay - deltay;
			}
			d = d + deltax + deltax;
 			motors (motor1_step, motor2_step, motor_delay);

// 			if (DEBUG) { debug_coordinates (x,y); }
		}
	} else {
		if (DEBUG) usart_puts ("y2<=y1\n");
		for (y = y1; y > y2; y--) {
 			motor1_step = motor_cw (motor1_step);
			if (d > 0) {
				x += i;
				if (i > 0) motor2_step = motor_ccw (motor2_step); else motor2_step = motor_cw (motor2_step);
				d = d - deltay - deltay;
			}
			d = d + deltax + deltax;
 			motors (motor1_step, motor2_step, motor_delay);

// 			if (DEBUG) { debug_coordinates (x,y); }
		}
	}
}

void plotter_move_pen_to_coordinates (int x1, int y1, int x2, int y2)
{
	// algorithm for drawing straight lines
	// it is almost exact copy from Wikipedia
	int absdeltax;
	int absdeltay;

	powersave_off ();

	absdeltax = 0;
	if (x1 > x2) absdeltax = x1 - x2;
	if (x1 < x2) absdeltax = x2 - x1;

	absdeltay = 0;
	if (y1 > y2) absdeltay = y1 - y2;
	if (y1 < y2) absdeltay = y2 - y1;

	if (absdeltax > absdeltay) {
		// absdeltax > absdeltay	- increment x, compute y
		plot_low (x1, y1, x2, y2);
	} else {
		// absdeltax < absdeltay	- increment y, compute x
		plot_high (x1, y1, x2, y2);
	}
}

void hpgl_command_in (void)
{
	int i;
	// IN = initialize, start a plotting job
	// move pen at left limit, to be sure where pen is (or use limit switch)
	powersave_off ();
	plotter_pen_up ();
	plotter_move_pen_to_coordinates (pen_x, pen_y, 0, 0);
	plotter_set_coordinates (0,0);
}

void hpgl_command_sp (int pen)
{
	// select pen
	// maybe plotter should pause, go to coordinates [0,0],
	// beep shortly, then wait until user will replace pen?
	beep (20, 50);
	delay (1000);
	beep (20, 90);
	delay (1000);
}

void hpgl_command_ro (int angle)
{
	// rotate = swap x and y coordinates and their limits
	// 0° and 180°	- portrait
	// 90° and 270°	- landscape
	switch (angle) {
		case 0:
		case 90:
		case 180:
		case 270:
			break;
		default:
			break;
	}
}

/*void hpgl_command_ma (int x, int y)
{
	// move absolute - it's command from MZXR language
}*/

/*void hpgl_command_mr (int x, int y)
{
	// move relative - it's command from MZXR language
}*/

void search_start_position (void)
{
	unsigned int counter;
	int i;
	
	// move pen at left limit, to be sure where pen is (or use limit switch)
	powersave_off ();
	plotter_pen_up ();
	
	// value 0x01 = switch is pressed
	PORTA |= 0x30;					// pullup for limit switch and second sensor whatever it will be
	
	if ((PINA & 0x10) == 0)
	{
		simplebeep ();
		counter = 0;
		limit_switch_found = 1;
		
		while ((PINA & 0x10) == 0)
		{
			// move pen one step to left side
			motor2_step = motor_cw (motor2_step);
			motors (motor1_step, motor2_step, MOTOR_DEFAULT_DELAY);

			// to not move pen indefinitelly
			if (counter > PAGEWIDTH_STEPS) break;
			counter++;
		}

		// and slowly move some steps back, only to switch not be pressed (too hard)
		counter = 0;
		while (1)
		{
			motor2_step = motor_ccw (motor2_step);
			motors (motor1_step, motor2_step, MOTOR_DEFAULT_DELAY*3);
			if ((counter > 20) && ((PINA & 0x10) == 0x10)) break;
			if (counter > 40) break;
			counter++;
		}
	}
	else
	{
		// if end switch is not detected, then move some steps to left and back, but nothing more
		// let it on user to place pen on propper place
		for (i = 0; i < 10; i++)
		{
			motor2_step = motor_cw (motor2_step);
			motors (motor1_step, motor2_step, MOTOR_DEFAULT_DELAY);
		}
		for (i = 0; i < 10; i++)
		{
			motor2_step = motor_ccw (motor2_step);
			motors (motor1_step, motor2_step, MOTOR_DEFAULT_DELAY);
		}
	}

	plotter_set_coordinates (0,0);
	beep (80, 20);
	beep (60, 40);
	beep (40, 60);
	beep (20, 80);
	L2_OFF;
}

//============================================================================

#define	MAX_PARAMETER_LENGTH	5	// numbers can be max 99999 (in reality max +-10800 because physical limits)
#define	MAX_PARAMETERS		5	// most commands have 2, max 4 parameters

typedef enum {GL_COMMAND, GL_NUMERIC_PARAMETER, MINIGRAF_COMPATIBLE} hpglstates;

int main(void)
{
	DDRA = 0b11001111;	// PA6..7 LED, PA4..5 sensors, PA0..3 motor_1
	DDRB = 0b00011111;	// ISP, PB4 speaker, PB0..3 motor_2
	DDRC = 0b00000000;	// 8 IO ports for anything, probably buttons
	DDRD = 0b11110110;	// PD7 debug, PD6 pen, PD5 motor 1 enable, PD4 motor 2 enable
				// PD0..PD3 RS232 / PD0 RxD (input), PD1 TxD (output), PD2 CTS (output), PD3 RTS (input)

	CTS_STOP;
	
	hpglstates hpglstate;

	int i;					// universal index
	unsigned char buttons = 0;		// buttons state
	unsigned char last_buttons = 0;
	unsigned char c;			// recieved character
	unsigned char character_counter = 0;
	unsigned char repeat = 0;		// repeat processing of character
	unsigned char manual_mode_motor = 0;	// true if motor should move in manual mode
	
	unsigned char is_online;
	unsigned char parameter_index;		// parameter counter
	char numeric_parameter_signum;		// 1/-1
	long int numeric_parameter_value;		// long ?
	int parameters;				// how many parameters to expect for current command
	int parameter[MAX_PARAMETERS];		// up too 5 16bit parameters are possible (some command like CI?)
	
	char command[] = "  ";			// empty command
	c = 0;					// no data
	
	motor1_step = 0;
	motor2_step = 0;
	motors (motor1_step, motor2_step, motor_delay);
	motors_on ();
	motor_delay = MOTOR_DEFAULT_DELAY;			// default delay between steps (it will be configurable)
	
	for (i = 0; i < MAX_PARAMETERS; i++) parameter[i] = 0;
	numeric_parameter_signum = 1;
	numeric_parameter_value = 0;

	hpglstate = GL_COMMAND;

	start_effect ();

	initialize_usart ();
	set_baudrate (9600);	// fast enough

	PORTC = 0xff;		// pullups for all buttons
	PORTA = 0x30;		// pullup for limit switch and second sensor whatever it will be

	delay (5);
	buttons = PINC;
	
	// PWM setup
        ICR1 = PWM_TOP;
        TCCR1A = 0xA2;		// COMA / COMB = clear on compare match, set on reset
				// FOCA / FOCB = 0
        TCCR1B = 0x19;		// WGM = 1110 = Fast PWM mode, ICR1 = TOP => frequency
				// CS = 001 = no prescaling clk/1

	OCR1A = PWM_PAPER_MOTOR;	// motor 1
	OCR1B = PWM_PEN_MOTOR;		// motor 2

	powersave_off ();
	search_start_position ();
	
	if ((buttons & 0x80) == 0)
	{
		hpglstate = MINIGRAF_COMPATIBLE;
		if (DEBUG) usart_puts ("Minigraf Aritma 0507 emulation\n");
		is_online = 0;
		last_buttons = buttons & 0b00111111;
		
		for (i = 0; i < 3; i++) {
			delay (10);
			beep (50,2);
			beep (50,30);
		}
	} else {
		// default
		usart_puts ("Ready\n");
		delay (5);
		sei ();
		CTS_RECIEVE;
		is_online = 1;
	}

	while (1)
	{
		if (powersave_counter > 0)
			powersave_counter--;
		else
			powersave_on ();
		
		buttons = PINC;
		
		if (hpglstate == MINIGRAF_COMPATIBLE) {
			// Aritma Minigraf 0507/0517 compatibility mode
			if ((buttons & 0xC0) == 0) {
				// delay is defined in MZXR
				move_pen_up (0);
			} else {
				// delay is defined in MZXR
				move_pen_down (0);
			}
			
			if ((buttons & 0b00111111) != (last_buttons & 0b00111111))
			{
				// motor 1 - paper
				// motor 2 - pen
				powersave_off ();
				powersave_counter = 65535;
				motor1_step = ((buttons >> 3) & 0b00000111);
				motor2_step = 7 - buttons & 0b00000111;
				last_buttons = buttons & 0b00111111;

				motors (motor1_step, motor2_step, MOTOR_DEFAULT_DELAY-2);
			}
		} else {
			if (buttons == 0xff) last_buttons = 0xff;
			
			// panel with buttons
			if (is_online)
			{
				// online mode - receiving data and drawing
				if ((buttons & 0x01) == 0 && (last_buttons & 0x01) != 0) {
					// left
					beep (10,10);
					last_buttons = buttons;
					delay (KEY_DELAY);
				}
				if ((buttons & 0x02) == 0 && (last_buttons & 0x02) != 0) {
					// up
					beep (10,10);
					last_buttons = buttons;
					delay (KEY_DELAY);
				}
				if ((buttons & 0x04) == 0 && (last_buttons & 0x04) != 0) {
					// down
					beep (10,10);
					last_buttons = buttons;
					delay (KEY_DELAY);
				}
				if ((buttons & 0x08) == 0 && (last_buttons & 0x08) != 0) {
					// right
					beep (10,10);
					last_buttons = buttons;
					delay (KEY_DELAY);
				}
				if ((buttons & 0x30) == 0) {
					// clear RS232 buffer / cancel print job
					plotter_pen_up ();
					if (usart_bytes > 0) {
						while (usart_bytes > 0) c = usart_getc ();
						simplebeep ();
					}
					plotter_move_pen_to_coordinates (pen_x, pen_y, 0, 0);
					plotter_set_coordinates (0,0);
					command [0] = ' ';
					command [1] = ' ';
					hpglstate = GL_COMMAND;
					parameter_index = 0;
				} else {
					if ((buttons & 0x10) == 0 && (last_buttons & 0x10) != 0)
					{
						// right
						beep (10,10);
						last_buttons = buttons;
						delay (KEY_DELAY);
					}
					if ((buttons & 0x20) == 0 && (last_buttons & 0x20) != 0)
					{
						// right
						beep (10,10);
						last_buttons = buttons;
						delay (KEY_DELAY);
					}
				}
				if ((buttons & 0x40) == 0 && (last_buttons & 0x40) != 0) {
					// switch to manual mode
					is_online = 0;
					last_buttons = buttons;
					usart_puts ("offline\n");
					beep (15,40);		// ?kHz
					beep (15,30);		// ?kHz
					beep (30,20);		// ?kHz
					delay (KEY_DELAY);
				}
			}
			else
			{
				// manual mode - moving pen and paper by buttons on panel
				manual_mode_motor = 0;
				
				if ((buttons & 0x01) == 0) {
					//left
					if ((PINA & 0x10) == 0x00)
					{
						motor2_step = motor_cw (motor2_step);
						manual_mode_motor = 1;
					} else {
						L2_ON;
						delay (motor_delay);
						L2_OFF;
					}
				}

				if ((buttons & 0x02) == 0) {
					//up
					motor1_step = motor_ccw (motor1_step);
					manual_mode_motor = 1;
				}
				
				if ((buttons & 0x04) == 0) {
					//down
					motor1_step = motor_cw (motor1_step);
					manual_mode_motor = 1;
				}
				
				if ((buttons & 0x08) == 0) {
					//right
					motor2_step = motor_ccw (motor2_step);
					manual_mode_motor = 1;
				}
				
				if ((buttons & 0x10) == 0) {
					// pen up
					plotter_pen_up ();
				}
				
				if ((buttons & 0x20) == 0) {
					// pen down
					plotter_pen_down ();
				}

				if ((buttons & 0x40) == 0 && (last_buttons & 0x40) != 0) {
					// mode
					is_online = 1;
					last_buttons = buttons;
					usart_puts ("online\n");
					beep (15,20);		// ?kHz
					beep (15,30);		// ?kHz
					beep (30,40);		// ?kHz
					delay (KEY_DELAY);
				}
				
				if (manual_mode_motor == 1) {
					powersave_off ();
					motors (motor1_step, motor2_step, motor_delay);
				}
			}

			/*if (((buttons & 0x01) == 0) && (!is_online || (last_buttons & 0x01) != 0)) {
				// left
				last_buttons = buttons;
				if (DEBUG) usart_puts ("B1\n");
				if (!is_online)
				{
					powersave_off ();
					if ((PINA & 0x10) == 0x00)
					{
						motor2_step = motor_cw (motor2_step);
						motors (motor1_step, motor2_step, motor_delay);
					} else {
						L2_ON;
						delay (motor_delay);
						L2_OFF;
					}
				} else {
					beep (20,10);
					delay (KEY_DELAY);
				}
			}

			if (((buttons & 0x02) == 0) && (!is_online || (last_buttons & 0x02) != 0)) {
				// up
				last_buttons = buttons;
				if (DEBUG) usart_puts ("B2\n");
				if (!is_online)
				{
					powersave_off ();
					motor1_step = motor_ccw (motor1_step);
					motors (motor1_step, motor2_step, motor_delay);
				} else {
					beep (20,10);
					delay (KEY_DELAY);
				}
			}
			
			if (((buttons & 0x04) == 0) && (!is_online || (last_buttons & 0x04) != 0)) {
				// down
				last_buttons = buttons;
				if (DEBUG) usart_puts ("B3\n");
				if (!is_online)
				{
					powersave_off ();
					motor1_step = motor_cw (motor1_step);
					motors (motor1_step, motor2_step, motor_delay);
				} else {
					beep (20,10);
					delay (KEY_DELAY);
				}
			}

			if (((buttons & 0x08) == 0) && (!is_online || (last_buttons & 0x08) != 0)) {
				// right
				last_buttons = buttons;
				if (DEBUG) usart_puts ("B4\n");
				if (!is_online)
				{
					powersave_off ();
					motor2_step = motor_ccw (motor2_step);
					motors (motor1_step, motor2_step, motor_delay);
				} else {
					beep (20,10);
					delay (KEY_DELAY);
				}
			}

			if (((buttons & 0x10) == 0) && ((last_buttons & 0x10) != 0)) {
				// toggle pen up/down
				last_buttons = buttons;
				if (DEBUG) usart_puts ("B5\n");
				if (!is_online)
				{
					plotter_pen_up ();
					delay (KEY_DELAY);
				} else {
					beep (20,10);
					delay (KEY_DELAY);
				}
			}

			if (((buttons & 0x20) == 0) && ((last_buttons & 0x20) != 0)) {
				// toggle speed
				last_buttons = buttons;
				if (DEBUG) usart_puts ("B6\n");
				if (!is_online)
				{
					plotter_pen_down ();
					delay (KEY_DELAY);
				} else {
					beep (20,10);
					delay (KEY_DELAY);
				}
			}*/

/*			if (((buttons & 0x40) == 0) && ((last_buttons & 0x40) != 0)) {
				// reset
				last_buttons = buttons;
				if (DEBUG) usart_puts ("MODE ");

				if (is_online)
				{
					is_online = 0;
					usart_puts ("offline\n");
					beep (20,40);		// ?kHz
					beep (20,30);		// ?kHz
					beep (50,20);		// ?kHz
				}
				else
				{
					is_online = 1;
					usart_puts ("online\n");
					beep (20,20);		// ?kHz
					beep (20,30);		// ?kHz
					beep (50,40);		// ?kHz
				}
				delay (KEY_DELAY);
			}*/

		
			// green status LED - online/offline
			if (is_online) L1_ON; else L1_OFF;
			// yellow status LED is for buffer overflow
		}

		if ((usart_bytes > 0) && (is_online > 0)) {
			switch (hpglstate) {
				case GL_COMMAND:
					// command can contain only uppercase letters
					switch (c)
					{
						case 'A'...'Z':
							if (character_counter == 0) {
								command [0] = c;
								character_counter = 1;
							} else {
								command [1] = c;
								
								if (strcmp (command, "IN") == 0) {
									// init
									hpgl_command_in ();
									if (DEBUG) debug_command ("cmd ", command, 1);
									
									// this command have no parameters in this implementation
									parameters = 0;
									character_counter = 0;
									hpglstate = GL_COMMAND;
								}
								if (strcmp (command, "OA") == 0) {
									// output actual positions
									if (DEBUG) debug_command ("cmd ", command, 1);
									if (DEBUG) {
										usart_puts ("pen position xy = [");
									}
									debug_int (pen_x);
									usart_puts (", ");
									debug_int (pen_y);
									usart_puts ("]\n");
									parameters = 0;
									character_counter = 0;
									hpglstate = GL_COMMAND;
								}
								if (strcmp (command, "OS") == 0) {
									// output status
									if (DEBUG) debug_command ("cmd ", command, 1);
									usart_puts ("Kapsa Robotics plotter 2019\n");
									parameters = 0;
									character_counter = 0;
									hpglstate = GL_COMMAND;
								}
								if (strcmp (command, "SP") == 0) {
									// set pen
									if (DEBUG) debug_command ("cmd ", command, 1);
									plotter_pen_up ();
									
									character_counter = 0;
									parameters = 1;				// expect one integer
									parameter_index = 0;
									numeric_parameter_signum = 1;
									numeric_parameter_value = 0;
									hpglstate = GL_NUMERIC_PARAMETER;
								}
								if (strcmp (command, "DF") == 0) {
									// set default value
									if (DEBUG) debug_command ("cmd ", command, 1);
								
									character_counter = 0;
									parameters = 2;				// expect two integers or nothing
									parameter_index = 0;
									numeric_parameter_signum = 1;
									numeric_parameter_value = 0;
									hpglstate = GL_NUMERIC_PARAMETER;
								}
								if ((command[0] == 'P') && ((command[1] == 'D') || (command[1] == 'U') || (command[1] == 'A') || (command[1] == 'R')))
								{
									// common part for commands PU, PD, PA, PR
									if (DEBUG) debug_command ("cmd ", command, 1);
								
									character_counter = 0;
									parameters = 2;				// expect two integers or nothing
									parameter_index = 0;
									numeric_parameter_signum = 1;
									numeric_parameter_value = 0;
									hpglstate = GL_NUMERIC_PARAMETER;
								}
								if ((character_counter > 0) && (hpglstate == GL_COMMAND)) {
									// unrecognized command
									if (DEBUG) debug_command ("cmd? ", command, 0);
									// use second letter as first and continue parsing
									command [0] = command [1];
									command [1] = ' ';
									character_counter = 1;
								}
								break;
							case '\n':
							case '\r':
								// ignore CR LF
								break;
							case ';':
							case ' ':
							case '\t':
								// reset command and ignore that characters
								command [0] = ' ';
								command [1] = ' ';
								character_counter = 0;
								break;
							default:
								// not letter, not whitespace
								if (DEBUG) debug_command ("no A-Z ", command, 0);
								command [0] = ' ';
								command [1] = ' ';
								character_counter = 0;
								break;
						}
					}
					break;
				case GL_NUMERIC_PARAMETER:
					switch (c) {
						case '-':
							numeric_parameter_signum = -1;
							if (DEBUG) { usart_putc ('m'); }
							break;
						case '0'...'9':
							if (character_counter < MAX_PARAMETER_LENGTH) {
								// just add next number into parameter
								if (DEBUG) { usart_putc ('n'); usart_putc (c); }
								numeric_parameter_value = numeric_parameter_value * 10;
								c -= 48;
								numeric_parameter_value += c;
								character_counter++;
							} else {
								// ERROR - parameter out of range, drop whole command
								if (DEBUG) usart_puts ("number out of range\n");
								hpglstate = GL_COMMAND;
								command [0] = ' ';
								command [1] = ' ';
								character_counter=0;
							}
							break;
						case ',':
						case ';':
							// separator, finish this numeric parameter
							if (numeric_parameter_value > 32767) numeric_parameter_value = 32767;
							if (numeric_parameter_value < -32768) numeric_parameter_value = -32768;
							
							if (DEBUG) { usart_putc ('s'); usart_putc (c); usart_putc ('\n'); }

							if (numeric_parameter_signum < 0) numeric_parameter_value = -numeric_parameter_value;
							parameter[parameter_index] = numeric_parameter_value;
							numeric_parameter_value = 0;
							if (DEBUG) {
								usart_puts ("param. ");
								debug_int (parameter[parameter_index]);
								usart_putc ('\n');
							}
							character_counter = 0;
							parameter_index++;
// 							if (DEBUG) {
// 								usart_puts ("parameter_index++ ");
// 								debug_int (parameter_index);
// 								usart_putc ('\n');
// 							}

							if (parameter_index == parameters || parameters == 0 || (command[0] == 'P' && c == ';'))
							{
								if (DEBUG) {
									usart_puts ("params. ");
									debug_int (parameters); usart_puts (", {");
									debug_int (parameter[0]); usart_puts (", ");
									debug_int (parameter[1]); usart_puts (", ");
									debug_int (parameter[2]); usart_puts (", ");
									debug_int (parameter[3]); usart_puts (", ");
									debug_int (parameter[4]); usart_puts ("}\n");
								}

								// expected number of parameters achieved, proceed command
								if (strcmp (command, "SP") == 0) {
									if (DEBUG) debug_int (parameter[0]);
									hpgl_command_sp (parameter[0]);
								}
								if (strcmp (command, "RO") == 0) {
									if (DEBUG) debug_int (parameter[0]);
									hpgl_command_ro (parameter[0]);
								}
								if (strcmp (command, "PD") == 0) {
									if (pen_status == UP) plotter_pen_down ();
									
									// if two parameters was readed, then change pen position
									if (parameter_index == 2) {
										// only positive coordinates are allowed
										if (parameter[0] < 0) parameter[0] = 0;
										if (parameter[1] < 0) parameter[1] = 0;
										if (DEBUG) debug_coordinates (parameter[0], parameter[1]);

										plotter_move_pen_to_coordinates (pen_x, pen_y, parameter[0], parameter[1]);
										plotter_set_coordinates (parameter[0],parameter[1]);
										parameter_index = 0;
									}
								}
								if (strcmp (command, "PU") == 0) {
									if (pen_status == DOWN) plotter_pen_up ();
									
									// if two parameters was readed, then change pen position
									if (parameter_index == 2) {
										// only positive coordinates are allowed
										if (parameter[0] < 0) parameter[0] = 0;
										if (parameter[1] < 0) parameter[1] = 0;
										if (DEBUG) debug_coordinates (parameter[0], parameter[1]);
										
										plotter_move_pen_to_coordinates (pen_x, pen_y, parameter[0], parameter[1]);
										plotter_set_coordinates (parameter[0],parameter[1]);
										parameter_index = 0;
									}
								}
								if (strcmp (command, "PA") == 0) {
									if (parameter_index == 2) {
										// only positive coordinates are allowed
										if (parameter[0] < 0) parameter[0] = 0;
										if (parameter[1] < 0) parameter[1] = 0;
										if (DEBUG) debug_coordinates (parameter[0], parameter[1]);
										
										plotter_move_pen_to_coordinates (pen_x, pen_y, parameter[0], parameter[1]);
										plotter_set_coordinates (parameter[0],parameter[1]);
										parameter_index = 0;
									}
								}
								if (strcmp (command, "PR") == 0) {
									if (parameter_index == 2) {
										// final coordinates must be positive, but relative movement can be negative
										if (pen_x + parameter[0] < 0) parameter[0] = pen_x;
										if (pen_y + parameter[1] < 0) parameter[1] = pen_y;
// 										if (pen_x + parameter[0] > 32768) parameter[0] = 32768 - pen_x;
// 										if (pen_y + parameter[1] > 32768) parameter[1] = 32768 - pen_y;
										if (DEBUG) debug_coordinates (parameter[0], parameter[1]);
										
										plotter_move_pen_to_coordinates (pen_x, pen_y, pen_x + parameter[0], pen_y + parameter[1]);
										plotter_set_coordinates (parameter[0],parameter[1]);
										parameter_index = 0;
									}
								}
								if (command[0] == 'P' && (c == ',' || c == ';'))
								{
									// reset used values for commands with more pairs of parameters
									for (i = 0; i < MAX_PARAMETERS; i++) parameter[i] = 0;
									numeric_parameter_signum = 1;
									numeric_parameter_value = 0;
								}
								if (strcmp (command, "DF") == 0) {
									// if two parameters was readed, then change pen position
									if (parameter_index == 2) {
										// only positive coordinates are allowed
										switch (parameter[0]) {
											case 'm':
												if (parameter[1] < 1) parameter[1] = 1;
												if (parameter[1] > 50) parameter[1] = 50;
												motor_delay = parameter[1];
												break;
										}
										parameter_index = 0;
									}
								}
							}
							if (c == ';') {
								// end of whole command and all groups of its parameters
								command [0] = ' ';
								command [1] = ' ';
								hpglstate = GL_COMMAND;
							}
							break;
						case 'A'...'Z':
							// end of command, but letter may be part of next command, don't drop it
							if (DEBUG) usart_puts ("missing ;\n");
							repeat = 1;
							hpglstate = GL_COMMAND;
							break;
						default:
							// unknown character... ignore it
							if (DEBUG) usart_putc ('?');
							break;
					}
					break;
				default:
					if (DEBUG) usart_puts ("unknown mode\n");
					break;
			}
			// if flag repeat is active, then repeat whole loop with same character, don't read new one
			// because in HPGL can be letter after numeric parameter which is part of following command
			if (repeat == 1)
				repeat = 0;		// reset flag
			else
			{
				c = usart_getc ();
				if (c < 32) c = ' ';
			}
		}
	}

}
