#include <avr/io.h>
#include <avr/interrupt.h>

#define DEBUG		1			// debug mode enable

#define F_OSC		16000000

#define L1_ON		PORTA |= 0x40		// SMD green LED on PCB
#define L1_OFF		PORTA &= ~0x40
#define L2_ON		PORTA |= 0x80		// SMD yellow LED on PCB
#define L2_OFF		PORTA &= ~0x80

#define LEDS_ON		PORTA |= 0xc0;		// both LEDs on
#define LEDS_OFF	PORTA &= ~0xc0;		// both LEDs off

#define SPK_ON		PORTB |= 0x10		// speaker
#define SPK_OFF		PORTB &= ~0x10

#define POC_ON		PORTD |= 0x80		// counter with display for debugging
#define POC_OFF		PORTD &= ~0x80

#define PEN_DOWN	PORTD |= 0x40		// plotter pen
#define PEN_UP		PORTD &= ~0x40

#define CTS_STOP	PORTD |= 0x04		// signal CTS to the computer L = recieving
#define CTS_RECIEVE	PORTD &= ~0x04

#define XON	17	// CTRL+Q / continue	// not used for now.... it wasn't reliable enough
#define XOFF	19	// CTRL+S / stop

//==== RUZNE =================================================================

int strcmp (const char * s1, const char * s2)
{
	const unsigned char *p1 = (const unsigned char *)s1;
	const unsigned char *p2 = (const unsigned char *)s2;

	while (*p1 && *p1 == *p2) ++p1, ++p2;

	return (*p1 > *p2) - (*p2  > *p1);
}

void delay (unsigned int ms)
{
	unsigned int i,j;
		for ( j=0; j<100; j++)
			for ( i=0; i<ms; i++);
}

void beepdelay (unsigned int ms)
{
	unsigned int i,j;
		for ( j=0; j<10; j++)
			for ( i=0; i<ms; i++);
}

void simplebeep (void)
{
	unsigned char n;

	for (n = 0; n < 2; n++)
	{
		SPK_ON;
		beepdelay (2);
		SPK_OFF;
		beepdelay (2);
	}
}

void beep (unsigned int cycles, unsigned int del)
{
	unsigned int n;

	for (n = 0; n < cycles; n++)
	{
		SPK_ON;
		beepdelay (del);
		SPK_OFF;
		beepdelay (del);
		
		// for 16MHz clock
		// 20 = 1.667kHz
		// 23 = 1.471kHz
		// 26 = 1.316kHz
		// 29 = 1.179kHz
	}
}

void debug_pulses (unsigned char pulses)
{
	// generate pulses on debug output
	unsigned char n;

	for (n = 0; n < pulses; n++)
	{
		POC_ON;
		beepdelay (2);
		POC_OFF;
		beepdelay (2);
	}
}

void start_effect ()
{
	LEDS_ON;
	beep (50,30);
	beep (50,10);
	delay (200);
	L1_OFF;
	delay (100);
	L2_OFF;
	delay (100);
}

//==== MOTORS ================================================================

const unsigned char motor_tab [] = {0x01, 0x05, 0x04, 0x06, 0x02, 0x0A, 0x08, 0x09};

//	driver	wire	port
//------------------------------------------
//	A	yellow	bit 0	enable A
//	B	red	bit 1	enable A
//	C	white	bit 2	enable B
//	D	blue	bit 3	enable B

//  0001  0x01
//  0101  0x05
//  0100  0x04
//  0110  0x06
//  0010  0x02
//  1010  0x0A
//  1000  0x08
//  1001  0x09

void motors (int step_m1, int step_m2)
{
	PORTA &= motor_tab [step_m1];
	PORTA |= motor_tab [step_m1];

	PORTB &= motor_tab [step_m2];
	PORTB |= motor_tab [step_m2];
}

unsigned char motor_cw (unsigned char step)
{
	if (step >= 7) step = 0; else step++;
	return (step);
}

unsigned char motor_ccw (unsigned char step)
{
	if (step == 0) step = 7; else step--;
	return (step);
}

void motor1_on (void)
{
	PORTD |= 0x10;
}

void motor1_off (void)
{
	PORTD &= ~0x10;
}

void motor2_on (void)
{
	PORTD |= 0x20;
}

void motor2_off (void)
{
	PORTD &= ~0x20;
}


//==== RS232 =================================================================

#define USART_MAX_BYTES		8
#define USART_BUFFER_SIZE	30

volatile unsigned char usart_buffer [USART_BUFFER_SIZE + 1];
volatile unsigned char usart_index_in;
volatile unsigned char usart_index_out;
volatile unsigned char usart_bytes;

void set_baudrate (unsigned long int new_baudrate)
{
	unsigned int ubrr = 0;

	// U2X = 1 (assynchronous double speed mode), default 9600bps, MCU clock 16MHz
	switch (new_baudrate)
	{
		case 2400:
			ubrr = 832;
			break;
		case 4800:
			ubrr = 416;
			break;
		default:
		case 9600:
			ubrr = 207;
			break;
		case 19200:
			ubrr = 103;
			break;
		case 57600:
			ubrr = 34;
			break;
		case 115200:
			ubrr = 16;
			break;
	}
	UBRRH = (unsigned char) (ubrr >> 8);
	UBRRL = (unsigned char) (ubrr);
}

int initialize_usart()
{
	unsigned char i;

	set_baudrate (9600);

	// clear buffer
	for (i = 0; i < USART_BUFFER_SIZE; i++) usart_buffer [i] = 0;

	usart_index_in = 0;
	usart_index_out = 0;
	usart_bytes = 0;

	// set registers
	UCSRA = 0x22;	// 0010 0010 (RXC, TXC, UDRE, FE, DOR, PE, U2X, MPCM)
	UCSRB = (1 << RXEN) | (1 << TXEN) | (1 << RXCIE);	// enable receiver and transmitter, enable RX interrupt
	UCSRC = (1 << URSEL) | (1 << USBS) | (3 << UCSZ0);      // 8 data, 2 stop bits
	// U2X manual page 147
}

void usart_putc (unsigned char c)
{
	// send one character - wait until register for sending is not free
	while(!(UCSRA & (1 << UDRE)));
	UDR = c;                // posli znak
}

void usart_puts (char *s)
{
	// send whole string of characters
	CTS_STOP;
	while (*s)
	{
		usart_putc(*s);
		s++;
	}
	CTS_RECIEVE;
}

unsigned char usart_getc(void)
{
	// get recieved character from USART register
	unsigned char byte;

	// get and return byte from buffer
	if (usart_bytes > 0)
	{
		byte = usart_buffer [usart_index_out];

		usart_index_out++;
		if (usart_index_out >= USART_BUFFER_SIZE) usart_index_out = 0;

		usart_bytes--;
		if (usart_bytes < USART_MAX_BYTES) CTS_RECIEVE;

		return (byte);
	}
	else
		return (0);
}

ISR (USART_RXC_vect)
{
	// interrupt after byte was recieved into USART
	unsigned char a;
	a = UDR;

	usart_buffer [usart_index_in] = UDR;
	usart_bytes++;

	if (usart_bytes > USART_MAX_BYTES) CTS_STOP;		// if buffer is almost full, stop sending device
	if (usart_bytes > USART_BUFFER_SIZE) simplebeep();	// this should never happen

	usart_index_in++;
	if (usart_index_in >= USART_BUFFER_SIZE) usart_index_in = 0;
}

//============================================================================
// paper size A4 210x297mm - printable area cca 190x270mm
// orig motors - 8 steps per mm - 1520 x 2160 steps (7600 x 10800 in HPGL coortinates)
// new motors - 20 steps per mm - 3800 x 5400 steps (7600 x 10800 in HPGL coortinates)
// for comparison - XY4160 have limits 7400x10400 in HPGL coordinates, and 10 motor steps per mm
// values needs to be revided by the real physical limits

#define	PLOTER_MAX_PEN_DOWN	1000;	// how long can be pen down (timeout)
#define	PLOTER_MAX_X		7600;	// HPGL limits determined by dimmensions of working area
#define	PLOTER_MAX_Y		10800;

typedef enum {UP, DOWN} penstates;

volatile penstates pen_status;
volatile unsigned int pen_timeout;	// will be decremented until 0, then pen will go up
volatile int pen_x;			// real absolute coordinates, how far from borders of working area pen really is
volatile int pen_y;			// (should it be in motor steps or HPGL units?)

void plotter_pen_up (void)
{
	PEN_UP;
	pen_status = UP;
	pen_timeout = 0;
}

void plotter_pen_down (void)
{
	PEN_DOWN;
	pen_status = DOWN;
	pen_timeout = PLOTER_MAX_PEN_DOWN;
}

void plotter_pen_refresh (void)
{
	pen_timeout = PLOTER_MAX_PEN_DOWN;
	if (pen_status == DOWN) PEN_DOWN; else PEN_UP;
}

void plotter_set_coordinates (unsigned int x, unsigned int y)
{
	pen_x = x;
	pen_y = y;
}

void plotter_move_pen_to_coordinates (unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	// algorithm for drawing straight lines
	unsigned int deltax;
	unsigned int deltay;

	deltax = 0;
	if (x1 > x2) deltax = x1 - x2;
	if (x1 < x2) deltax = x2 - x1;

	deltay = 0;
	if (y1 > y2) deltax = y1 - y2;
	if (y1 < y2) deltax = y2 - y1;
	
	if ((deltax > 0) || (deltay > 0))
	{
		if (deltax == 0)
		{
			// line parallel to y axis
		}

		if (deltay == 0)
		{
			// line parallel to x axis
		}

		if (deltax == deltay)
		{
			// line in angle 45°
		}
		else
			if ((deltax > 0) && (deltay > 0))
			{
				if (deltax > deltay)
				{
				}
				else
				{
				}
			}
	}
	// else nothing to do
}

void hpgl_command_in (void)
{
	// IN = initialize, start a plotting job
	plotter_pen_up ();
	plotter_set_coordinates (0,0);
}

void hpgl_command_sp (void)
{
	// select pen, maybe plotter should pause, go to [0,0] and beep shortly,
	// then wait until user will replace pen and press button?
}

void hpgl_command_ma (void)
{
	// move absolute
}

void hpgl_command_mr (void)
{
	// move relative
}

//============================================================================

#define	MAX_PARAMETER_LENGTH	5	// numbers can be max 99999 (in reality max +-10800 because physical limits)

typedef enum {GL_COMMAND, GL_NUMERIC_PARAMETER} hpglstates;

int main(void)
{
	DDRA = 0b11001111;	// PA6..7 LED, PA4..5 sensors, PA0..3 motor_1
	DDRB = 0b00011111;	// ISP, PB4 speaker, PB0..3 motor_2
	DDRC = 0b00000000;	// 8 IO ports for anything, probably buttons
	DDRD = 0b11110110;	// PD7 debug, PD6 pen, PD5 motor 1 enable, PD4 motor 2 enable
				// PD0..PD3 RS232 / PD0 RxD, PD1 TxD, PD2 RTS, PD3 CTS

	CTS_STOP;
	
	hpglstates hpglstate;

	unsigned char buttons = 0;		// buttons state
	unsigned char last_buttons = 0;
	unsigned char c;			// recieved character
	unsigned char character_counter = 0;
	unsigned char repeat = 0;		// repeat processing of character
	
	unsigned char is_online;
	unsigned char parameter_index;		// parameter counter
	char numeric_parameter_signum;		// 1/-1
	long int numeric_parameter_value;
	unsigned char parameters;		// how many parameters to expect for current command
	int parameter1;
	int parameter2;
	
	char command[] = "  ";			// empty command
	c = 0;					// no data

	hpglstate = GL_COMMAND;

	start_effect ();

	initialize_usart ();
	set_baudrate (115200);

	PORTC = 0xff;		// pullups for all buttons
	PORTA = 0x30;		// pullup for limit switch and second sensor whatever it will be

	delay (5);
	usart_puts ("Ready\n");
	delay (5);

	sei ();
	CTS_RECIEVE;
	
	is_online = 1;

	while (1)
	{
		buttons = PINC;
		
		if (buttons == 0xff) last_buttons = 0xff;
		
		if (((buttons & 0x01) == 0) && ((last_buttons & 0x01) != 0)) {
			// inicializace plotru? vlozeni papiru?
			last_buttons = buttons;
			if (DEBUG) usart_puts ("B1\n");
			
// 			if (is_online)
// 			{
// 				is_online = 0;
// 				usart_puts ("offline\n");
// 			}
// 			else
// 			{
// 				is_online = 1;
// 				usart_puts ("online\n");
// 			}
			beep (20,10);
			delay (20);
		}

/*		if (((buttons & 0x02) == 0) && ((last_buttons & 0x02) != 0)) {
			last_buttons = buttons;
			if (DEBUG) usart_puts ("button 2\n");
			beep (20,10);
			delay (20);
		}
		
		if (((buttons & 0x04) == 0) && ((last_buttons & 0x04) != 0)) {
			last_buttons = buttons;
			if (DEBUG) usart_puts ("button 3\n");
			beep (20,10);
			delay (20);
		}

		if (((buttons & 0x08) == 0) && ((last_buttons & 0x08) != 0)) {
			last_buttons = buttons;
			if (DEBUG) usart_puts ("button 4\n");
			beep (20,10);
			delay (20);
		}*/
		
		// status LED
		if (is_online) L1_ON; else L1_OFF;
		if (pen_status == DOWN) L2_ON; else L2_OFF;

		if ((usart_bytes > 0) && (is_online)) {
			switch (hpglstate) {
				case GL_COMMAND:
					// command can contain only uppercase letters
					if ((c >= 'A') && (c <= 'Z')) {
						if (character_counter == 0) {
							command [0] = c;
							character_counter = 1;
						} else {
							command [1] = c;
							
							if (strcmp (command, "IN") == 0) {
								hpgl_command_in ();
								if (DEBUG){
									debug_pulses (1);
									usart_puts ("cmd in\n");
									beep (10,10);
								}
								
								// this command have no parameters in this implementation
								character_counter = 0;
								hpglstate = GL_COMMAND;
							}
							if (strcmp (command, "SP") == 0) {
								// set pen
								// move pen at left limit, to be sure where pen is (or use limit switch)
								if (DEBUG){
									debug_pulses (1);
									usart_puts ("cmd sp\n");
									beep (10,10);
								}
								
								plotter_pen_up ();
								
								character_counter = 0;
								parameters = 1;				// expect one decimal number
								parameter_index = 0;
								numeric_parameter_value = 0;
								hpglstate = GL_NUMERIC_PARAMETER;
							}
							if (strcmp (command, "PD") == 0) {
								// pen down
								if (DEBUG){
									debug_pulses (1);
									usart_puts ("cmd pd\n");
									beep (10,10);
								}

								if (pen_status == UP) plotter_pen_down ();
								
								character_counter = 0;
								parameters = 2;
								parameter_index = 0;
								numeric_parameter_value = 0;
								hpglstate = GL_NUMERIC_PARAMETER;
							}
							if (strcmp (command, "PU") == 0) {
								// pen up
								if (DEBUG){
									debug_pulses (1);
									usart_puts ("cmd pu\n");
									beep (10,10);
								}
								
								if (pen_status == DOWN) plotter_pen_up ();
								
								character_counter = 0;
								parameters = 2;
								parameter_index = 0;
								numeric_parameter_value = 0;
								hpglstate = GL_NUMERIC_PARAMETER;
							}
							if (strcmp (command, "PA") == 0) {
								// plot absolute
								if (DEBUG){
									debug_pulses (1);
									usart_puts ("cmd pa\n");
									beep (10,10);
								}
								
								plotter_pen_refresh ();
								
								character_counter = 0;
								parameters = 2;
								parameter_index = 0;
								numeric_parameter_value = 0;
								hpglstate = GL_NUMERIC_PARAMETER;
							}
							if (strcmp (command, "PR") == 0) {
								// plot relative
								if (DEBUG){
									debug_pulses (1);
									usart_puts ("cmd pr\n");
									beep (10,10);
								}
								
								plotter_pen_refresh ();
								
								character_counter = 0;
								parameters = 2;
								parameter_index = 0;
								numeric_parameter_value = 0;
								hpglstate = GL_NUMERIC_PARAMETER;
							}
							
							if ((character_counter > 0) && (hpglstate == GL_COMMAND)) {
								// unrecognized command
								if (DEBUG) usart_puts ("cmd ?\n");
								// use second letter as first and continue parsing
								command [0] = command [1];
								command [1] = ' ';
 								character_counter = 1;
							}
						}
					}
					else
					{
						if (c != ' ' && c != 13 && c != 10 && c != ',' && c != ';') {
							if (DEBUG) usart_puts ("not A-Z\n");
						}
						character_counter = 0;
						hpglstate = GL_COMMAND;
					}
					break;
				case GL_NUMERIC_PARAMETER:
					if (c >= '0' && c <= '9')
					{
						if (character_counter < MAX_PARAMETER_LENGTH) {
							// just add next number into parameter
							numeric_parameter_value = numeric_parameter_value * 10 + c - 48;
							if (DEBUG) usart_puts ("adding\n");
							character_counter++;
						} else {
							// too long parameter, cannot add another number, ignore whole command
							if (DEBUG) usart_puts ("numeric parameter is too long!\n");
							hpglstate = GL_COMMAND;
							character_counter=0;
						}
					}
					// current character is not number, then parameter is finished or empty
					if (character_counter > 0) {
						// parameter finished
						if (DEBUG) usart_puts ("not number\n");
						
						if (parameter_index == 0) {
							parameter_index++;
							// X
							parameter1 = numeric_parameter_value;
						} else {
							if ((parameter_index + 1) >= parameters)
								if (DEBUG) usart_puts ("too much parameters\n");
							parameter_index = 0;
							// Y
							parameter2 = numeric_parameter_value;

							// HERE interprett commands
							// printf ("ma%g,%g\n", (double)(hpgl_x/40.0), (double)(hpgl_y/40.0));
						}
						
						numeric_parameter_value = 0;
						character_counter = 0;
						if (c >= 'A' && c <= 'Z')
						{
							if (DEBUG) usart_puts ("letter after parameter\n");
							repeat = 1;
							hpglstate = GL_COMMAND;
						}
						if (c == ';') {
							if (DEBUG) usart_puts ("end of command\n");
							hpglstate = GL_COMMAND;
						}
					} else {
						// parameter is empty
						if (DEBUG) usart_puts ("empty parameter\n");
						numeric_parameter_value = 0;
						character_counter = 0;
						if (c >= 'A' && c <= 'Z') {
							repeat = 1;
							hpglstate = GL_COMMAND;
						}
						if (c == ';') hpglstate = GL_COMMAND;
					}
					break;
				default:
					if (DEBUG) usart_puts ("unknown mode\n");
					break;
			}
			// if flag repeat is active, then repeat whole loop with same character, don't read new one
			// because in HPGL can be letter after numeric parametter which is part of following command
			if (repeat == 1)
				repeat = 0;		// reset flag
			else
			{
				c = usart_getc ();
// 				if ((c == 10) || (c == 13)) c = ' ';
				if (c < 32) c = ' ';
			}
		}
	}
}
